#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Smart, Asynchronous Link Tester with Database backend (SALTED)
~~~~~~~~~~~~~~~~~~~~~
Source: https://github.com/RuedigerVoigt/salted
(c) 2020-2021 by Rüdiger Voigt
Released under the Apache License 2.0
"""
import sys

if sys.version_info.major != 3 or sys.version_info.minor < 8:
    raise RuntimeError('You need at least Python 3.8 to run salted. ' +
                       'You are running: ' + str(sys.version_info.major) +
                       '.' + str(sys.version_info.minor))

from salted.__main__ import Salted

NAME = "salted"
__version__ = "0.6.0"
__author__ = "Rüdiger Voigt"



#!/usr/bin/python3
# -*- coding: utf-8 -*-

"""
Input Handler for salted
~~~~~~~~~~~~~~~~~~~~~
Source: https://github.com/RuedigerVoigt/salted
(c) 2020-2021: Released under the Apache License 2.0
"""

from collections import Counter
import logging
import pathlib
from typing import List
import re
import urllib.parse


from bs4 import BeautifulSoup  # type: ignore
import userprovided
from tqdm.asyncio import tqdm  # type: ignore

from salted import database_io


class InputHandler:
    """Methods to find files and the hyperlinks inside them."""

    def __init__(self,
                 db: database_io.DatabaseIO):
        self.db = db
        self.cnt: Counter = Counter()

        self.pattern_latex_url = re.compile(
            r"\\url\{(?P<url>[^{]*?)\}",
            flags=re.MULTILINE | re.IGNORECASE)

        self.pattern_latex_href = re.compile(
            r"\\href(\[.*\]){0,1}\{(?P<url>[^}]*)\}\{(?P<linktext>[^}]*?)\}",
            flags=re.MULTILINE | re.IGNORECASE)

        self.pattern_md_link = re.compile(
            r"\[(?P<linktext>[^\[]*)\]\((?P<url>[^\)]*?)[\s\)]+",
            flags=re.MULTILINE | re.IGNORECASE)

        self.pattern_md_link_pointy = re.compile(
            r"<(?P<url>[^>]*?)>",
            flags=re.MULTILINE | re.IGNORECASE)

    @staticmethod
    def find_files_by_extensions(
            path_to_base_folder: pathlib.Path,
            suffixes: set = {".htm", ".html", '.md', '.tex'}
                                 ) -> List[pathlib.Path]:
        """Find all files with specific file type suffixes in the base folder
           and its subfolders. If no file suffix is specified, this will look
           for all file formats supported by salted."""
        files_to_check = []
        path_to_check = pathlib.Path(path_to_base_folder)
        all_files = path_to_check.glob('**/*')
        for candidate in all_files:
            if candidate.suffix in suffixes:
                files_to_check.append(candidate.resolve())
        logging.debug('Found %s files', len(files_to_check))
        return files_to_check

    def find_html_files(self,
                        path_to_base_folder: pathlib.Path
                        ) -> List[pathlib.Path]:
        """Find all HTML files in the base folder and its subfolders."""
        logging.info("Looking for HTML files...")
        suffixes = {".htm", ".html"}
        return self.find_files_by_extensions(path_to_base_folder, suffixes)

    def find_markdown_files(self,
                            path_to_base_folder: pathlib.Path
                            ) -> List[pathlib.Path]:
        """Find all markdown files in the base folder and its subfolders."""
        logging.info("Looking for Markdown files...")
        suffixes = {".md"}
        return self.find_files_by_extensions(path_to_base_folder, suffixes)

    def find_tex_files(self,
                       path_to_base_folder: pathlib.Path
                       ) -> List[pathlib.Path]:
        """Find all .tex files in the base folder and its subfolders."""
        logging.info("Looking for TeX files...")
        suffixes = {".tex"}
        return self.find_files_by_extensions(path_to_base_folder, suffixes)

    def extract_links_from_html(self,
                                file_content: str) -> List:
        """Extract all links from a HTML file."""
        matches = []
        soup = BeautifulSoup(file_content, 'html.parser')
        for link in soup.find_all('a'):
            matches.append([link.get('href'), link.text])
        return matches

    def extract_links_from_markdown(self,
                                    file_content: str) -> List:
        """Extract all links from a Markdown file.
        Returns a list of lists: [[url, linktext], [url, linktext]]"""
        matches = []
        md_links_in_file = re.findall(self.pattern_md_link, file_content)
        for match in md_links_in_file:
            matches.append([match[1], match[0]])
        pointy_links_in_file = re.findall(self.pattern_md_link_pointy,
                                          file_content)
        for url in pointy_links_in_file:
            matches.append([url, url])
        return matches

    def extract_links_from_tex(self,
                               file_content: str) -> List:
        """Extract all links from a .tex file.
        Returns a list of lists: [[url, linktext], [url, linktext]]"""
        matches = []
        # extract class \href{url}{text} links
        href_in_file = re.findall(self.pattern_latex_href, file_content)
        for match in href_in_file:
            # The RegEx returns the optinal Element as first element.
            # (Empty, but still in the return if it is not in the string.)
            matches.append([match[1], match[2]])
        # extract \url{url} links
        url_in_file = re.findall(self.pattern_latex_url, file_content)
        for url in url_in_file:
            matches.append([url, url])
        return matches

    def scan_files_for_links(self,
                             files_to_check: List[pathlib.Path]) -> None:
        """Scan each file within a list of paths for hyperlinks and write
           those to the SQLite database. """
        if not files_to_check:
            logging.warning('No files to check')
            return None

        # Reset counter as check_links might be used multiple times and this
        # should be per run:
        self.cnt['links_found'] = 0

        print("Scanning files for links:")
        for file_path in tqdm(files_to_check):
            content = ''
            with open(file_path, 'r') as code:
                content = code.read()
            if file_path.suffix in {".htm", ".html"}:
                extracted = self.extract_links_from_html(content)
            elif file_path.suffix in {".md"}:
                extracted = self.extract_links_from_markdown(content)
            elif file_path.suffix in {".tex"}:
                extracted = self.extract_links_from_tex(content)
            else:
                raise RuntimeError('Invalid extension. Should never happen.')

            links_found = []
            for link in extracted:
                url = link[0]
                linktext = link[1]
                if url.startswith('http'):
                    # It may be that multiple links point to the same resource.
                    # Normalizing them means they only need to be tested once.
                    # The non-normalized version is stored anyway, because if
                    # the link is broken, that version is used to show the
                    # user the broken links on a specific page.
                    normalized_url = userprovided.url.normalize_url(url)
                    parsed_url = urllib.parse.urlparse(url)
                    links_found.append([str(file_path),
                                        parsed_url.hostname,
                                        url,
                                        normalized_url,
                                        linktext])
                    self.cnt['links_found'] += 1

            # Push the found links once for each file instead for all files
            # at once. The latter would kill performance for large document
            # collections.
            if links_found:
                self.db.save_found_links(links_found)

        return None



#!/usr/bin/python3
# -*- coding: utf-8 -*-

"""
Handle the database and cache for salted.
~~~~~~~~~~~~~~~~~~~~~
Source: https://github.com/RuedigerVoigt/salted
(c) 2020-2021: Released under the Apache License 2.0
"""

import logging
import pathlib
import sqlite3
from typing import Optional, Union


class DatabaseIO:
    """Create the database schema (tables, views, ...), handle the cache file
       and log the crawler's results to sqlite. """

    def __init__(self,
                 dont_check_again_within_hours: int,
                 cache_file: Union[pathlib.Path, str] = None):

        self.dont_check_again_within_hours = dont_check_again_within_hours

        self.cache_file_path: Optional[pathlib.Path] = None
        if cache_file:
            self.cache_file_path = pathlib.Path(cache_file).resolve()
            logging.debug('Absolute path to cache file: %s',
                          self.cache_file_path)
            self.check_cache_file_path()

        # Prepare to create a sqlite database in memory
        # Not actually created here, because users might reuse the object
        # for multiple checks on different document sets. Therefore the
        # check_links() function in __main__.py will call init_in_memory_db
        # and tear_down_in_memory_db().
        self.conn = sqlite3.connect(
            ':memory:',
            isolation_level=None  # reenable autocommit
            )
        self.cursor = self.conn.cursor()

    def get_cursor(self) -> sqlite3.Cursor:
        """Returns a valid cursor. is used by the report generator that
           directly accesses the database outside this class.
           If user reuse the object a new database is created and in that case
           the old cursor is invalid."""
        return self.cursor

    def reinitialize_in_memory_db(self):
        """If there is already an in memory instance, close it. Then create
           and initialize a new in memory instance of the database by
           creating its schema and loading the disk cache (if available)."""
        logging.debug('Reinitialize in memory database.')
        self.tear_down_in_memory_db()
        self.conn = sqlite3.connect(
            ':memory:',
            isolation_level=None  # reenable autocommit
            )
        self.cursor = self.conn.cursor()
        self.create_schema()
        self.load_disk_cache()

    def tear_down_in_memory_db(self):
        """If there is an active in memory instance, close the connection.
           All data not stored elsewhere will be lost."""
        if self.conn:
            logging.debug("tear down in memory instance")
            self.conn.close()

    def check_cache_file_path(self) -> None:
        """Check if the given path is valid in order to fail if it is not
           before the linkcheck runs.
           Raise ValueError if the path is a directory or if parent
           folders do not exists."""

        if not self.cache_file_path:
            raise RuntimeError('check_cache_file path called without path set')

        if self.cache_file_path.exists() and self.cache_file_path.is_file():
            return

        # Established that the file does not exist, but check if it can
        # exist before returning False.
        if not self.cache_file_path.parent.is_dir():
            raise ValueError('Incorrect path to cache_file. ' +
                             'Parameter cache_file must be the path to ' +
                             'a file and parent directories must exist.')
        if self.cache_file_path.is_dir():
            raise ValueError('Parameter cache_file is a directory, ' +
                             'but must include file name!')

    def create_schema(self) -> None:
        """Create the SQLite database schema. """
        self.cursor.execute('''
            CREATE TABLE links (
            filePath text,
            hostname text,
            url text,
            normalizedUrl text,
            linktext text);''')
        self.cursor.execute('''
            CREATE TABLE errors (
            normalizedUrl text,
            error integer);''')
        self.cursor.execute('''
            CREATE TABLE permanentRedirects (
                normalizedUrl text,
                error integer);''')
        self.cursor.execute('''
            CREATE TABLE exceptions (
            normalizedUrl text,
            reason text);''')
        self.cursor.execute('''
            CREATE TABLE IF NOT EXISTS validUrls (
            normalizedUrl text,
            lastValid integer);''')

        logging.debug("Created database schema.")

    def load_disk_cache(self) -> None:
        """If there is a cache file open it, read the valid URLs and
           load them into the in-memory instance of sqlite."""

        if not self.cache_file_path:
            return

        try:
            disk_cache = sqlite3.connect(
                self.cache_file_path,
                isolation_level=None  # reenable autocommit
            )
            disk_cache_cursor = disk_cache.cursor()
            disk_cache_cursor.execute('''
                SELECT
                normalizedUrl, lastValid
                FROM validUrls
                WHERE lastValid > (strftime('%s','now') - (? * 3600));''',
                [self.dont_check_again_within_hours])
            valid_urls = disk_cache_cursor.fetchall()
            disk_cache.close()

            if valid_urls:
                self.cursor.executemany('''
                    INSERT INTO validUrls
                    (normalizedUrl, lastValid)
                    VALUES (?, ?);''', valid_urls)

        except Exception:
            logging.debug('No cache file or could not read it.', exc_info=True)
            pass

    def generate_indices(self) -> None:
        """Add indices to the in memory database. This is not done on creation
           for performance reasons."""
        # While adding links to the database the index is not needed,
        # but would be updated with every insert. It is faster to create it
        # once the table has it contents.
        self.cursor.execute('''
            CREATE INDEX IF NOT EXISTS index_timestamp
            ON validUrls (lastValid);''')
        self.cursor.execute('''
            CREATE INDEX IF NOT EXISTS index_normalized_url
            ON links (normalizedUrl);''')

    def generate_db_views(self) -> None:
        """ Generate Views for Analytics and Output Generating."""
        # Separate function to execute after all links have been checked
        # and the respective tables are stable."""
        self.cursor.execute('''
            CREATE VIEW IF NOT EXISTS v_errorCountByFile AS
            SELECT COUNT(*) AS numErrors, filePath
            FROM links
            WHERE normalizedUrl IN (
            SELECT normalizedUrl FROM errors
            ) GROUP BY filePath;''')
        self.cursor.execute('''
            CREATE VIEW IF NOT EXISTS v_redirectCountByFile AS
            SELECT COUNT(*) AS numRedirects, filePath
            FROM links
            WHERE normalizedUrl IN (
            SELECT normalizedUrl FROM permanentRedirects
            ) GROUP BY filePath;''')
        self.cursor.execute('''
            CREATE VIEW IF NOT EXISTS v_exceptionCountByFile AS
            SELECT COUNT(*) AS numExceptions, filePath
            FROM links
            WHERE normalizedUrl IN (
            SELECT normalizedUrl FROM exceptions
            ) GROUP BY filePath;''')

        self.cursor.execute('''
            CREATE VIEW IF NOT EXISTS v_errorsByFile AS
            SELECT links.filePath,
            links.url,
            links.linktext,
            errors.error AS httpCode
            FROM links
            INNER JOIN errors
            ON links.normalizedUrl = errors.normalizedUrl;''')

        self.cursor.execute('''
            CREATE VIEW IF NOT EXISTS v_redirectsByFile AS
            SELECT links.filePath,
            links.url,
            links.linktext,
            permanentRedirects.error AS httpCode
            FROM links
            INNER JOIN permanentRedirects
            ON links.normalizedUrl = permanentRedirects.normalizedUrl;''')

        self.cursor.execute('''
            CREATE VIEW IF NOT EXISTS v_exceptionsByFile AS
            SELECT links.filePath,
            links.url,
            links.linktext,
            exceptions.reason
            FROM links
            INNER JOIN exceptions
            ON links.normalizedUrl = exceptions.normalizedUrl;''')

        logging.debug('Created Views for analytics and output generating.')

    def save_found_links(self,
                         links_found: list) -> None:
        """Save the links found into the memory database."""
        if not links_found:
            logging.debug('No links in this file to save them.')
        else:
            self.cursor.executemany('''
            INSERT INTO links
            (filePath, hostname, url, normalizedUrl, linktext)
            VALUES(?, ?, ?, ?, ?);''', links_found)

    def urls_to_check(self) -> Optional[list]:
        """Return a list of all distinct URLs to check."""
        self.cursor.execute('SELECT DISTINCT normalizedUrl FROM links;')
        return self.cursor.fetchall()

    def log_url_is_fine(self,
                        url: str) -> None:
        """If a request to an URL returns a HTTP status code that indicates a
           working hyperlink, note that with a timestamp."""
        self.cursor.execute('''
            INSERT INTO validUrls
            (normalizedUrl, lastValid)
            VALUES (?, strftime('%s','now'));''', [url])

    def log_error(self,
                  url: str,
                  error_code: int) -> None:
        """An error is logged for HTTP status codes that indicate a permanently
           broken link like '404 - File Not found' or '410 Gone'."""
        self.cursor.execute('INSERT INTO errors VALUES (?, ?);',
                            [url, error_code])

    def log_redirect(self,
                     url: str,
                     code: int) -> None:
        """Logs permanent redirects. Those links *should* be fixed. """
        self.cursor.execute('''INSERT INTO permanentRedirects
                               (normalizedUrl, error)
                               VALUES (?, ?);''', [url, code])

    def log_exception(self,
                      url: str,
                      exception_str: str) -> None:
        """An exception is logged if it was not possible to check
           a specific URL."""
        self.cursor.execute('''INSERT INTO exceptions VALUES (?, ?);''',
                            [url, exception_str])

    def del_links_that_can_be_skipped(self) -> int:
        """If links from a non-expired cache have been read, try to eliminate
           them in the list of URLs to check.
           Return the absolute number of (non-normalized) URLs to check."""

        self.cursor.execute('SELECT COUNT(*) FROM links;')
        num_links_before = self.cursor.fetchone()[0]

        self.cursor.execute('''DELETE FROM links
                            WHERE normalizedUrl IN (
                            SELECT normalizedUrl FROM validUrls);''')

        self.cursor.execute('SELECT COUNT(*) FROM links;')
        num_links_after = self.cursor.fetchone()[0]

        if num_links_before > num_links_after:
            msg = (f"Tests for {num_links_before - num_links_after} " +
                   "hyperlinks skipped - they checked out valid within " +
                   f"the last {self.dont_check_again_within_hours} " +
                   "hours.\n" +
                   f"{num_links_after} unique links have to be tested.")
            logging.info(msg)
        return num_links_after

    def count_errors(self) -> int:
        """Return the number of errors. """
        self.cursor.execute('SELECT COUNT(*) FROM errors;')
        return self.cursor.fetchone()[0]

    def overwrite_cache_file(self) -> None:
        """Write the current in-memory database into a file.
           Overwrite any file in the given path."""

        self.cache_file_path.unlink(missing_ok=True)  # type: ignore[union-attr]

        if self.cache_file_path:
            new_cache_file = sqlite3.connect(self.cache_file_path)
            with new_cache_file:
                self.conn.backup(new_cache_file, name='main')
            new_cache_file.close()



#!/usr/bin/python3
# -*- coding: utf-8 -*-

"""
Smart, Asynchronous Link Tester with Database backend (SALTED)
~~~~~~~~~~~~~~~~~~~~~
Source: https://github.com/RuedigerVoigt/salted
(c) 2020: Released under the Apache License 2.0
"""

import asyncio
from collections import Counter
import logging
import pathlib
import time
from typing import Optional, Union

from tqdm.asyncio import tqdm  # type: ignore
import userprovided

from salted import database_io
from salted import input_handler
from salted import network_interaction
from salted import report_generator


class Salted:
    """Main class. Creates the other Objects, starts workers,
       collects results and starts the report of results. """

    version = '0.6.0'  # released: Jan 08, 2021

    def __init__(self,
                 cache_file: Union[pathlib.Path, str],
                 workers: Union[int, str] = 'automatic',
                 timeout_sec: int = 5,
                 dont_check_again_within_hours: int = 24,
                 raise_for_dead_links: bool = False,
                 user_agent: str = f"salted/{version}") -> None:

        self.num_workers = workers
        self.timeout = int(timeout_sec)

        userprovided.parameters.enforce_boolean(
            raise_for_dead_links,
            'raise_for_dead_links')
        self.raise_for_dead_links = raise_for_dead_links

        self.user_agent = user_agent

        self.db = database_io.DatabaseIO(
            dont_check_again_within_hours,
            cache_file)
        self.file_io = input_handler.InputHandler(self.db)

        self.display_result = report_generator.ReportGenerator(
            self.db)

        self.cnt: Counter = Counter()
        self.num_checks = 0
        self.pbar_links: tqdm = None

    def __recommend_num_workers(self) -> int:
        """If the number of workers is set to 'automatic', this returns an
           estimate an appropriate number of async workers to use - based on
           the number of hyperlinks to check.
           If the user provided a specific number of workers, that will be
           returned instead."""

        if self.num_workers == 'automatic':
            if self.num_checks < 1:
                raise ValueError

            if self.num_checks < 25:
                recommendation = 4
            elif self.num_checks < 100:
                recommendation = 12
            elif self.num_checks > 99:
                recommendation = 32
        else:
            recommendation = int(self.num_workers)
        # Set the logging message here to flush the cache. Cannot use
        # flush() as it is unknow which or hwo many logging method are used.
        logging.info("%s unique hyperlinks to check. Using %s workers.",
                     self.num_checks, recommendation)
        return recommendation

    async def __worker(self,
                       name: str,
                       queue: asyncio.Queue) -> None:
        while True:
            url = await queue.get()
            await self.network.check_url(url, 'head')
            self.pbar_links.update(1)
            queue.task_done()

    async def __distribute_work(self,
                                urls_to_check: list) -> None:
        """Start a queue and spawn workers to work in parallel."""
        queue: asyncio.Queue = asyncio.Queue()
        for entry in urls_to_check:
            queue.put_nowait(entry[0])

        # initialize here as there has to exist an event loop
        self.network = network_interaction.NetworkInteraction(
            self.db,
            self.timeout,
            self.user_agent)

        tasks = []
        for i in range(int(self.num_workers)):
            task = asyncio.create_task(self.__worker(f'worker-{i}', queue))
            tasks.append(task)
        await queue.join()

        # Cancel worker tasks.
        for task in tasks:
            task.cancel()

        # Close aiohttp session
        await self.network.close_session()

        # Wait until all worker tasks are cancelled.
        await asyncio.gather(*tasks, return_exceptions=True)

    def check_links(self,
                    path_to_base_folder: Union[str, pathlib.Path],
                    template_searchpath: str = 'salted/templates',
                    template_name: str = 'default.cli.jinja',
                    write_to: Union[str, pathlib.Path] = 'cli',
                    base_url: Optional[str] = None) -> None:
        """Check all links found in files within the provided folder
           and its subfolders."""
        start_time = time.monotonic()

        # check_links might be reused with the same salted object. Therefore
        # the database has to reinitialized to remove data like exceptions 
        # et cetera from previous runs. however this loads the disk cache.
        self.db.reinitialize_in_memory_db()

        # Expand path as otherwise a relative path will not be rewritten
        # in output:
        path_to_base_folder = pathlib.Path(path_to_base_folder).resolve()
        logging.info('Base folder: %s', path_to_base_folder)

        # Remove trailing slash in base URL if there is one:
        if base_url:
            base_url = base_url.rstrip('/')

        files_to_check = self.file_io.find_files_by_extensions(
            path_to_base_folder)
        if files_to_check:
            self.file_io.scan_files_for_links(files_to_check)
            num_links_to_check = self.db.del_links_that_can_be_skipped()
            if num_links_to_check == 0:
                msg = ("Nothing to do after skipping cached results." +
                       "All hyperlinks are considered valid.")
                logging.info(msg)
                return
        else:
            logging.warning(
                "No supported files in this folder or its subfolders.")
            return

        urls_to_check = self.db.urls_to_check()

        if not urls_to_check:
            logging.warning("Found no hyperlinks to check!")
            return

        # Class level because this number is needed to initialize the progress
        # bar and will be logged in some messages:
        self.num_checks = len(urls_to_check)

        # Set of number of workers here instead of __distribute_work as
        # otherwise the logging message will force the progress bar to repaint.
        self.num_workers = self.__recommend_num_workers()

        self.pbar_links = tqdm(total=self.num_checks)

        # ##### START ASYNCHRONOUS CODE #####

        asyncio.run(self.__distribute_work(urls_to_check))

        # ##### END ASYNCHRONOUS CODE #####
        self.db.generate_indices()
        self.db.generate_db_views()
        self.pbar_links.close()

        runtime_check = time.monotonic() - start_time

        self.display_result.generate_report(
            statistics={
                'num_links': self.file_io.cnt['links_found'],
                'num_checked': self.num_checks,
                'time_to_check': (round(runtime_check)),
                'checks_per_second': (
                    round(self.num_checks / runtime_check, 2)),
                'num_fine': self.network.cnt['fine'],
                'needed_full_request': self.network.cnt['neededFullRequest'],
                'percentage_full_request': (
                    round(self.network.cnt['neededFullRequest'] /
                          self.num_checks * 100, 2))
                          },
            template={
                'searchpath': template_searchpath,
                'name': template_name,
                'foldername_to_replace': str(path_to_base_folder),
                'base_url': base_url},
            write_to=write_to,
            replace_path_by_url={
                'path_to_be_replaced': str(path_to_base_folder),
                'replace_with_url': base_url
            })
        if self.raise_for_dead_links:
            if self.db.count_errors() > 0:
                raise Exception("Found dead URLs")
        self.db.overwrite_cache_file()
        self.db.tear_down_in_memory_db()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="salted",
    version="0.6.0",
    author="Rüdiger Voigt",
    author_email="projects@ruediger-voigt.eu",
    description="Smart, Asynchronous Link Tester with Database backend",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/RuedigerVoigt/salted",
    package_data={
        "salted": ["py.typed", "templates/*.jinja"]},
    include_package_data=True,
    packages=setuptools.find_packages(),
    python_requires=">=3.8",
    install_requires=["aiodns>=2.0.0",
                      "aiohttp>=3.7.3",
                      "beautifulsoup4>=4.9.3",
                      "cchardet>=2.1.7",
                      "jinja2>=2.11.2",
                      "lxml>=4.6.2",
                      "tqdm>=4.54.1",
                      "userprovided>=0.8.0"],
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        #"Development Status :: 5 - Production/Stable",
        "Development Status :: 4 - Beta",
        "Framework :: AsyncIO",
        "Intended Audience :: Developers",
        "Intended Audience :: System Administrators",
        "Topic :: Text Processing :: Markup :: HTML",
        "Topic :: Internet :: WWW/HTTP :: Indexing/Search",
        "Topic :: Internet :: WWW/HTTP :: Site Management",
        "Topic :: Internet :: WWW/HTTP :: Site Management :: Link Checking"
    ],
)



#!/usr/bin/python3
# -*- coding: utf-8 -*-

"""
Network interactions for salted.
~~~~~~~~~~~~~~~~~~~~~
Source: https://github.com/RuedigerVoigt/salted
(c) 2020: Released under the Apache License 2.0
"""
import asyncio
from collections import Counter
import logging

import aiohttp

from salted import database_io


class NetworkInteraction:
    """Interacts with the network to check hyperlinks."""

    def __init__(self,
                 db: database_io.DatabaseIO,
                 timeout_sec: int,
                 user_agent: str) -> None:
        self.db = db
        self.timeout_sec = timeout_sec

        self.headers: dict = dict()
        if user_agent:
            self.headers = {'User-Agent': user_agent}

        self.session = aiohttp.ClientSession(loop=asyncio.get_running_loop())

        self.cnt: Counter = Counter()

    async def close_session(self):
        """Close the session object once it is no longer needed"""
        if self.session:
            await self.session.close()

    async def head_request(self,
                           url: str) -> int:
        """The HTTP HEAD method requests the headers, but not the body of
           a page. Requesting this way reduces load on the server and
           reduces network traffic."""
        async with self.session.get(url,
                                    headers=self.headers,
                                    raise_for_status=False,
                                    timeout=self.timeout_sec) as response:
            return response.status

    async def full_request(self,
                           url: str) -> int:
        """ Some servers do not understand or block a HTTP HEAD request.
            In those cases this function can try a full request.
            This carries the risk of encountering very large pages.
            Therefore the read is limited."""

        async with self.session.get(url,
                                    headers=self.headers,
                                    raise_for_status=False,
                                    timeout=self.timeout_sec) as response:
            await response.content.read(100)

        return response.status

    async def check_url(self,
                        url: str,
                        request_type: str) -> None:
        """Check the URL by using a HTTP HEAD request (or if necessary a full
           request with limited data read) to check the link and log the result
           to the database. """
        # pylint: disable=too-many-branches
        try:
            if request_type == 'head':
                response_code = await self.head_request(url)
            elif request_type == 'full':
                response_code = await self.full_request(url)
                self.cnt['neededFullRequest'] += 1

            if response_code in (200, 302, 303, 307):
                self.cnt['fine'] += 1
                self.db.log_url_is_fine(url)
            elif response_code in (301, 308):
                self.db.log_redirect(url, response_code)
            elif response_code == 403:
                if request_type == 'head':
                    await self.check_url(url, 'full')
                else:
                    self.db.log_error(url, 403)
            elif response_code in (404, 410):
                self.db.log_error(url, response_code)
            elif response_code == 429:
                self.db.log_exception(url, 'Rate Limit (429)')
            else:
                if request_type == 'head':
                    await self.check_url(url, 'full')
                else:
                    self.db.log_exception(url, f"Other ({response_code})")
        except asyncio.TimeoutError:
            self.db.log_exception(url, 'Timeout')
        except aiohttp.client_exceptions.ClientConnectorError:
            self.db.log_exception(url, 'ClientConnectorError')
        except aiohttp.client_exceptions.ClientResponseError:
            self.db.log_exception(url, 'ClientResponseError')
        except aiohttp.client_exceptions.ClientOSError:
            self.db.log_exception(url, 'ClientOSError')
        except aiohttp.client_exceptions.ServerDisconnectedError:
            self.db.log_exception(url, 'Server disconnected')
        except Exception:
            logging.exception('Exception. URL %s', url,  exc_info=True)
            raise



#!/usr/bin/python3
# -*- coding: utf-8 -*-

"""
Report generator for salted
~~~~~~~~~~~~~~~~~~~~~
Source: https://github.com/RuedigerVoigt/salted
(c) 2020: Released under the Apache License 2.0
"""
import logging
import pathlib
from typing import Optional, Union

from jinja2 import Environment, FileSystemLoader, PackageLoader

from salted import database_io


class ReportGenerator:
    """Generates a report about broken links and redirects for display.
       Reports can be styled using Jinja2 templates."""

    def __init__(self,
                 db_object: database_io.DatabaseIO,
                 show_redirects: bool = True,
                 show_exceptions: bool = True):
        self.db = db_object
        self.show_redirects = show_redirects
        self.show_exceptions = show_exceptions
        self.replace_path_by_url: Optional[dict] = None

    def __list_errors(self,
                      error_code: int) -> list:
        """Return a list of normalized URLs that yield a specific
           error code (from the HTTP status codes)."""
        cursor = self.db.get_cursor()
        cursor.execute('''SELECT normalizedUrl
                               FROM errors
                               WHERE error = ?;''', [error_code])
        urls_with_error = cursor.fetchall()
        if urls_with_error:
            return urls_with_error
        return list()

    def rewrite_path(self,
                     path_to_rewrite: str) -> str:
        """In a given path, replace the path to the folder
           with the base URL."""
        # Silence mypy index error, because this assures the values
        # are available:
        if not self.replace_path_by_url['path_to_be_replaced']:  # type: ignore
            raise ValueError('Cannot replace in URL not knowing what.')
        if not self.replace_path_by_url['replace_with_url']:  # type: ignore
            raise ValueError('Cannot replace in URL not knowing with what.')

        return path_to_rewrite.replace(
            self.replace_path_by_url['path_to_be_replaced'],  # type: ignore
            self.replace_path_by_url['replace_with_url'],  # type: ignore
            1)

    def generate_error_list(self) -> Optional[list]:
        """If the crawl found hyperlinks that yield permanent errors, return
           a list of dictionaries containing the file path, the number of
           permanent errors in that file, and a list of the actual errors."""
        cursor = self.db.get_cursor()
        result = list()
        cursor.execute(
            '''SELECT filePath, numErrors
               FROM v_errorCountByFile
               ORDER BY numErrors DESC, filePath ASC;''')
        pages_w_permanent_errors = cursor.fetchall()
        if not pages_w_permanent_errors:
            return None
        for file_path, num_errors in pages_w_permanent_errors:
            # The url as in the code, not the normalized version used to check.
            cursor.execute('''
                SELECT url, linktext, httpCode
                FROM v_errorsByFile
                WHERE filePath = ?;''', [file_path])
            defects = cursor.fetchall()
            if self.replace_path_by_url:
                file_path = self.rewrite_path(file_path)
            result.append({'path': file_path,
                           'num_errors': num_errors,
                           'defects': defects})
        return result

    def generate_redirect_list(self) -> Optional[list]:
        """If the crawl found hyperlinks that yield permanent redirects, return
           a list of dictionaries containing the file path, the number of
           permanent redirects in that file, and a list of the redirects."""
        cursor = self.db.get_cursor()
        result = list()
        cursor.execute(
            '''SELECT filePath, numRedirects
                FROM v_redirectCountByFile
                ORDER BY numRedirects DESC, filePath ASC;''')
        pages_w_redirects = cursor.fetchall()
        if not pages_w_redirects:
            return None
        for file_path, num_redirects in pages_w_redirects:
            # The url as in the code, not the normalized version used to check.
            cursor.execute('''
                SELECT url, linktext, httpCode
                FROM v_redirectsByFile
                WHERE filePath = ?;''', [file_path])
            redirects = cursor.fetchall()
            if self.replace_path_by_url:
                file_path = self.rewrite_path(file_path)
            result.append({'path': file_path,
                           'num_redirects': num_redirects,
                           'redirects': redirects})
        return result

    def generate_exception_list(self) -> Optional[list]:
        """If it was not possible due to exceptions to check or more links,
           return a list of dictionaries containing the file path, the number
           of exception causing links in that file, and a list of the
           actual exceptions."""
        cursor = self.db.get_cursor()
        result = list()
        cursor.execute(
            '''SELECT filePath, numExceptions
                FROM v_exceptionCountByFile
                ORDER BY numExceptions DESC, filePath ASC;''')
        pages_w_exceptions = cursor.fetchall()
        if not pages_w_exceptions:
            return None
        for file_path, num_exceptions in pages_w_exceptions:
            # The url as in the code, not the normalized version used to check.
            cursor.execute('''
                SELECT url, linktext, reason
                FROM v_exceptionsByFile
                WHERE filePath = ?;''', [file_path])
            exceptions = cursor.fetchall()
            if self.replace_path_by_url:
                file_path = self.rewrite_path(file_path)
            result.append({'path': file_path,
                           'num_exceptions': num_exceptions,
                           'exceptions': exceptions})
        return result

    def generate_report(self,
                        statistics: dict,
                        template: dict,
                        write_to: Union[str, pathlib.Path],
                        replace_path_by_url: dict = None
                        ) -> None:
        """Generate the report for the user with a Jinja2 template.
           Either display it at the command line interface or write
           it to a file. """
        cursor = self.db.get_cursor()
        # The base URL is always given. Invalidate the parameter if no
        # replacement is provided.
        if not replace_path_by_url['replace_with_url']:  # type: ignore[index]
            replace_path_by_url = None
        else:
            self.replace_path_by_url = replace_path_by_url

        permanent_errors = self.generate_error_list()

        permanent_redirects: Optional[list] = None
        if self.show_redirects:
            permanent_redirects = self.generate_redirect_list()

        crawl_exceptions: Optional[list] = None
        if self.show_exceptions:
            crawl_exceptions = self.generate_exception_list()

        rendered_report = ''

        if template['name'] in ('default.cli.jinja', 'default.md.jinja'):
            # built-in template
            jinja_env = Environment(loader=PackageLoader('salted', 'templates'))
            builtin_template = jinja_env.get_template(template['name'])
            rendered_report = builtin_template.render(
                statistics=statistics,
                permanent=permanent_errors,
                redirects=permanent_redirects,
                exceptions=crawl_exceptions)
        else:
            # external template from file system
            jinja_env = Environment(loader=FileSystemLoader(
                searchpath=template['searchpath']))
            user_template = jinja_env.get_template(template['name'])
            rendered_report = user_template.render(
                statistics=statistics,
                permanent=permanent_errors,
                redirects=permanent_redirects,
                exceptions=crawl_exceptions)

        if write_to == 'cli':
            print(rendered_report)
            return
        try:
            with open(write_to, 'w') as file:
                file.write(rendered_report)
            logging.info("Wrote report to file: %s",
                         pathlib.Path(write_to).resolve())
        except Exception:
            logging.exception('Exception while writing to file!',
                              exc_info=True)
            raise
